"""
    QuaO Project device_factory.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from ..enum.provider_type import ProviderType
from ..model.device.aws_braket_device import AwsBraketDevice
from ..model.device.ibm_cloud_device import IbmCloudDevice
from ..model.device.ibm_quantum_device import IbmQuantumDevice
from ..model.device.quao_device import QuaoDevice
from ..model.provider.provider import Provider


class DeviceFactory:
    @staticmethod
    def create_device(provider: Provider, device_specification: str, authentication: dict, sdk: str):
        """

        @param provider:
        @param device_specification:
        @param authentication:
        @return:
        """

        provider_type = provider.get_provider_type().value

        if ProviderType.QUAO_QUANTUM_SIMULATOR.value.__eq__(provider_type):
            return QuaoDevice(provider, device_specification, sdk)

        if ProviderType.IBM_QUANTUM.value.__eq__(provider_type):
            return IbmQuantumDevice(provider, device_specification)

        if ProviderType.IBM_CLOUD.value.__eq__(provider_type):
            return IbmCloudDevice(provider, device_specification)

        if ProviderType.AWS_BRAKET.value.__eq__(provider_type):
            return AwsBraketDevice(provider,
                                   device_specification,
                                   authentication.get('bucketName'),
                                   authentication.get('prefix'))

        raise Exception("Unsupported device!")



"""
    QuaO Project job_response.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from ...enum.media_type import MediaType
from ...enum.job_status import JobStatus


class JobResponse(object):
    def __init__(
            self,
            provider_job_id: str = "",
            job_status: str = "",
            job_result=None,
            content_type=None,
            job_histogram=None,
            user_identity="",
            user_token=""
    ):
        self.provider_job_id = provider_job_id if provider_job_id else ""
        self.job_status = job_status if job_status else JobStatus.ERROR.value
        self.job_result = job_result if job_result else None
        self.content_type = content_type if content_type else MediaType.ALL_TYPE.value
        self.job_histogram = job_histogram if job_histogram else None
        self.user_identity = user_identity
        self.user_token = user_token



"""
    QuaO Project ibm_quantum_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

from qiskit import transpile

from ...model.device.qiskit_device import QiskitDevice
from ...util.json_parser_util import JsonParserUtils
from ...config.logging_config import *


class IbmQuantumDevice(QiskitDevice):

    def _is_simulator(self) -> bool:
        return self.device.configuration().simulator

    def _parse_job_result(self, job_result) -> dict:
        return JsonParserUtils.parse(job_result.to_dict())

    def _create_job(self, circuit, shots):
        logger.debug('Create Ibm Quantum job with {0} shots'.format(shots))
        transpile_circuit = transpile(circuit, self.device)

        return self.device.run(transpile_circuit, shots=shots)



"""
    QuaO Project logging_config.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
# import logging

from loguru import logger

logger.add( format="{level} : {time} : {message}: {process}",  level='DEBUG')

# logging.basicConfig(format='%(asctime)s - %(message)s', level=logging.DEBUG)



"""
    QuaO Project ibm_quantum_provider.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from qiskit_ibm_provider import IBMProvider

from ...enum.provider_type import ProviderType
from ...model.provider.provider import Provider
from ...config.logging_config import *


class IbmQuantumProvider(Provider):
    def __init__(self, api_token):
        super().__init__(ProviderType.IBM_QUANTUM)
        self.api_token = api_token

    def get_backend(self, device_specification: str):
        """

        @param device_specification:
        """

        provider = self.collect_providers()

        return provider.get_backend(device_specification)

    def collect_providers(self):
        """

        @return:
        """

        logger.debug('Connect to Ibm Quantum provider')
        return IBMProvider(token=self.api_token)



"""
    QuaO Project ibm_cloud_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from qiskit_ibm_runtime import Options, Session, Sampler

from ...model.device.qiskit_device import QiskitDevice
from ...util.json_parser_util import JsonParserUtils
from ...config.logging_config import *


class IbmCloudDevice(QiskitDevice):
    def _is_simulator(self) -> bool:
        return self.device.configuration().simulator

    def _parse_job_result(self, job_result) -> dict:
        return JsonParserUtils.parse(job_result.__dict__)

    def _create_job(self, circuit, shots):
        logger.debug('Create Ibm Cloud job with {0} shots'.format(shots))
        options = Options(optimization_level=1)
        options.execution.shots = shots

        with Session(service=self.provider.collect_providers(), backend=self.device) as session:
            sampler = Sampler(session=session, options=options)
            job = sampler.run(circuits=circuit)

            return job



"""
    QuaO Project ProviderType.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from enum import Enum


class ProviderType(Enum):
    QUAO_QUANTUM_SIMULATOR = 'QUAO_QUANTUM_SIMULATOR'
    IBM_QUANTUM = 'IBM_QUANTUM'
    IBM_CLOUD = 'IBM_CLOUD'
    AWS_BRAKET = 'AWS_BRAKET'



"""
    QuaO Project json_parser_util.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from datetime import datetime
from enum import Enum

import numpy as np


class JsonParserUtils:

    @staticmethod
    def parse(unparsed_input: dict) -> dict:
        data_holder = {}

        for key, value in unparsed_input.items():
            if isinstance(value, (set, tuple)):
                value = list(value)
            if isinstance(value, np.ndarray):
                value = value.tolist()
            if isinstance(value, (complex, datetime)):
                value = value.__str__()

            if isinstance(value, dict):
                data_holder[key] = JsonParserUtils.parse(value)
            elif isinstance(value, (str, int, float, bool, Enum)) or value is None:
                data_holder[key] = value
            elif isinstance(value, (list,)):
                data_list_holder = []
                for data in value:
                    data = JsonParserUtils.__resolve_type(data)
                    if isinstance(data, dict):
                        data_list_holder.append(JsonParserUtils.parse(data))
                    else:
                        data_list_holder.append(data)
                data_holder[key] = data_list_holder
            else:
                data_holder[key] = JsonParserUtils.parse(value.__dict__)

        return data_holder

    @staticmethod
    def __resolve_type(element):
        if isinstance(element, (complex, datetime)):
            return element.__str__()
        if isinstance(element, (dict, str, int, float, bool, Enum)) or element is None:
            return element
        if isinstance(element, np.ndarray):
            return JsonParserUtils.__resolve_type(element.tolist())
        if isinstance(element, (list,)):
            data_holder = []
            for data in element:
                data_holder.append(JsonParserUtils.__resolve_type(data))
            return data_holder
        return element.__dict__



"""
    QuaO Project request_data.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""


class RequestData:

    def __init__(self, event):
        json_data = event.json()
        self.data = json_data
        self.input = json_data.get("input")
        self.shots = json_data.get("shots")
        self.device_id = json_data.get("deviceId")
        self.sdk = json_data.get("sdk").lower() if json_data.get("sdk") else None
        self.server_url = json_data.get("serverUrl")
        self.circuit_export_url = json_data.get("circuitExportUrl")
        self.user_token = json_data.get("userToken")
        self.user_identity = json_data.get("userIdentity")
